import dbus
from advertisement import Advertisement
from service import Application, Service, Characteristic, Descriptor

# Constants
GATT_CHRC_IFACE = "org.bluez.GattCharacteristic1"
SSID_NAMES_CHARACTERISTIC_UUID = "00000002-710e-4a5b-8d75-3e5b444bc3cf"
SSID_INFO_CHARACTERISTIC_UUID = "00000003-710e-4a5b-8d75-3e5b444bc3cf"
NOTIFY_TIMEOUT = 5000

class SSIDAdvertisement(Advertisement):
    def __init__(self, index):
        Advertisement.__init__(self, index, "peripheral")
        self.add_local_name("RoboBin-BT")
        self.include_tx_power = True

class SSIDService(Service):
    SSID_SERVICE_UUID = "00000001-710e-4a5b-8d75-3e5b444bc3cf"

    def __init__(self, index):
        Service.__init__(self, index, self.SSID_SERVICE_UUID, True)
        self.ssid_names = ["HomeNetwork", "OfficeWiFi", "GuestWiFi"]  # Example SSID list
        self.selectedSSID = ""
        self.password = ""
        self.add_characteristic(SSIDListCharacteristic(self))
        self.add_characteristic(SSIDInfoCharacteristic(self))

    def get_ssid_names(self):
        """Returns the SSID list as a string with newline separation."""
        return "\n".join(self.ssid_names)

    def set_ssid_info(self, ssidInfo):
        """Sets the SSID and password."""
        self.selectedSSID, self.password = ssidInfo.split(",")

class SSIDListCharacteristic(Characteristic):
    """A read-only characteristic that returns a list of SSIDs."""
    def __init__(self, service):
        Characteristic.__init__(
            self, SSID_NAMES_CHARACTERISTIC_UUID, ["read"], service)
        self.add_descriptor(SSIDListDescriptor(self))

    def ReadValue(self, options):
        """Returns the list of SSIDs joined by newline characters."""
        value = []
        ssid_names_str = self.service.get_ssid_names()
        
        for c in ssid_names_str:
            value.append(dbus.Byte(c.encode()))
    
        return value

class SSIDListDescriptor(Descriptor):
    TEMP_DESCRIPTOR_UUID = "2901"
    TEMP_DESCRIPTOR_VALUE = "SSID List"

    def __init__(self, characteristic):
        Descriptor.__init__(self, self.TEMP_DESCRIPTOR_UUID, ["read"], characteristic)

    def ReadValue(self, options):
        value = []
        desc = self.TEMP_DESCRIPTOR_VALUE
        for c in desc:
            value.append(dbus.Byte(c.encode()))
        return value

class SSIDInfoCharacteristic(Characteristic):
    """A read/write characteristic for sending an SSID and password."""
    def __init__(self, service):
        Characteristic.__init__(self, SSID_INFO_CHARACTERISTIC_UUID, ["read", "write", "notify"], service) #Adding "notify" to enable notifications
        self.add_descriptor(SSIDDescriptor(self))
        self.notifying = False #To track if a client is receiving notifications

    def ReadValue(self, options):
        """Returns the last sent SSID and password (if any)."""
        value = []
        ssid_info_str = (self.service.selectedSSID + ":" + self.service.password) or "No SSID info set"

        for c in ssid_info_str:
            value.append(dbus.Byte(c.encode()))
        
        return value

    def WriteValue(self, value, options):
        """Writes an SSID and password sent from the client."""
        ssid_info = "".join([chr(b) for b in value])  # Convert bytes back to string
        self.service.set_ssid_info(ssid_info)
        print(f"Received SSID and password: {ssid_info}")

    def StartNotify(self):
        """Called when a client subscribes to notifications."""
        if self.notifying:
            return
        
        self.notifying = True
        # If needed, start sending notifications, e.g., a periodic update like the CPU temperature example
        # self.add_timeout(NOTIFY_TIMEOUT, self.some_notify_callback)

    def StopNotify(self):
        """Called when a client unsubscribes from notifications."""
        self.notifying = False

class SSIDDescriptor(Descriptor):
    """Descriptor providing additional information about the SSID characteristics."""
    SSID_DESCRIPTOR_UUID = "2901"
    SSID_DESCRIPTOR_VALUE = "SSID name and password"

    def __init__(self, characteristic):
        Descriptor.__init__(self, self.SSID_DESCRIPTOR_UUID, ["read"], characteristic)

    def ReadValue(self, options):
        value = []
        desc = self.SSID_DESCRIPTOR_VALUE
        for c in desc:
            value.append(dbus.Byte(c.encode()))
        return value

# Main Application
app = Application()
app.add_service(SSIDService(0))
app.register()

adv = SSIDAdvertisement(0)
adv.register()

try:
    app.run()
except KeyboardInterrupt:
    app.quit()
